;;; rpm-spec-mode.el --- RPM spec file editing commands for XEmacs

;; Copyright (C) 1997,98 Stig Bjrlykke, <stigb@tihlde.hist.no>

;; Authors:  Stig Bjrlykke, <stigb@tihlde.hist.no> and
;;           Steve Sanbeg, <sanbeg@dset.com>
;; Keywords: unix, languages
;; Version:  0.08a

;; This file is not yet part of FSF Emacs or XEmacs.

;; Emacs/XEmacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; Emacs/XEmacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with Emacs/XEmacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
;; MA 02111-1307, USA.

;;; Synched up with:  both in FSF Emacs and XEmacs.

;;; Thanx:

;;   to Tore Olsen <toreo@tihlde.hist.no> for some general fixes. 

;;; ToDo:

;; - different buffernames for different rpm builds.
;; - rewrite function names and shortcuts.
;; - autofill changelog entries.
;; - customize rpm-tags-list and rpm-group-tags-list.
;; - get values from `rpm --showrc'.
;; - ssh/rsh for compile.
;; - finish integrating the new navigation functions in with existing stuff. 
;; - use a single prefix conistently (internal)

;;; Commentary:

;; This mode is used for editing spec files used for building RPM
;; packages.

;; Put this in your .emacs file to enable autoloading of rpm-spec-mode,
;; and auto-recognition of ".spec" files.
;;
;;  (autoload 'rpm-spec-mode "rpm-spec-mode.el" "RPM spec mode." t)
;;  (setq auto-mode-alist (append '(("\\.spec" . rpm-spec-mode))
;;                                auto-mode-alist))

;;------------------------------------------------------------

;;; Code:

(defgroup rpm-spec nil
  "RPM Spec mode with XEmacs enhancements."
  :prefix "rpm-spec"
  :group 'languages)

(defcustom rpm-spec-add-attr nil
  "Add %attr entry for filelistings or not."
  :type 'boolean
  :group 'rpm-spec)

(defcustom rpm-spec-short-circuit nil
  "Skip straight to specified stage.
(ie, skip all stages leading up to the specified stage).  Only valid
in \"%build\" and \"%install\" stage."
  :type 'boolean
  :group 'rpm-spec)

(defcustom rpm-spec-timecheck "0"
  "Set the \"timecheck\" age (0 to disable).
The timecheck value expresses, in seconds, the maximum age of a file
being packaged.  Warnings will be printed for all files beyond the
timecheck age."
  :type 'integer
  :group 'rpm-spec)

(defcustom rpm-spec-clean nil
  "Remove the build tree after the packages are made."
  :type 'boolean
  :group 'rpm-spec)

(defcustom rpm-spec-test nil
  "Do not execute any build stages.  Useful for testing out spec files."
  :type 'boolean
  :group 'rpm-spec)

(defcustom rpm-spec-sign-pgp nil
  "Embed a PGP signature in the package.
This signature can be used to verify the integrity and the origin of
the package."
  :type 'boolean
  :group 'rpm-spec)

(defcustom rpm-initialize-sections t
  "Automatically add empty section headings to new spec files"
  :type 'boolean
  :group 'rpm-spec)

(defgroup rpm-spec-faces nil
  "Font lock faces for RPM Spec mode."
  :group 'rpm-spec
  :group 'faces)

;;------------------------------------------------------------
;; variables used by navigation functions.
(defconst rpm-sections 
  '("preamble" "description" "prep" "setup" "build" "install" "clean"
    "changelog" "files")
  "Partial list of section names.")
(defconst rpm-scripts 
  '("pre" "post" "preun" "postun" "trigger" "triggerun" "triggerpostun")
  "List of rpm scripts")
(defconst rpm-section-seperate "^%\\(\\w+\\)\\s-")


;;(defconst rpm-section-regexp 
;;  "^%\\(\\(description\\)\\|\\(prep\\)\\|\\(build\\)\\|\\(install\\)\\|\\(files\\)\\|\\(clean\\)\\|\\(pre\\|post\\(un\\)?\\)\\|\\(trigger\\(post\\)?\\([iu]n\\)?\\)\\)\\b"
;;  "Regular expression to match beginning of a section.")

;; I added "package" to be recognized too, because is is kind of preamble 
;; rpm-current-section returns "preamble" instead "package" /klakier
(defconst rpm-section-regexp 
  "^%\\(\\(package\\)\\|\\(description\\)\\|\\(prep\\)\\|\\(build\\)\\|\\(install\\)\\|\\(files\\)\\|\\(clean\\)\\|\\(pre\\|post\\(un\\)?\\)\\|\\(trigger\\(post\\)?\\([iu]n\\)?\\)\\)\\b"
  "Regular expression to match beginning of a section.")

;;------------------------------------------------------------

(defface rpm-spec-tag-face
    '(( ((class color) (background light)) (:foreground "Blue3") )
      ( ((class color) (background dark)) (:foreground "Blue3") ))
  "*The face used for tags."
  :group 'rpm-spec-faces)

(defface rpm-spec-macro-face
    '(( ((class color) (background light)) (:foreground "orange3") )
      ( ((class color) (background dark)) (:foreground "orange3") ))
  "*The face used for macros."
  :group 'rpm-spec-faces)

(defface rpm-spec-doc-face
    '(( ((class color) (background light)) (:foreground "#6920ac") )
      ( ((class color) (background dark)) (:foreground "#6920ac") ))
  "*The face used for document files."
  :group 'rpm-spec-faces)

(defface rpm-spec-dir-face
    '(( ((class color) (background light)) (:foreground "green4") )
      ( ((class color) (background dark)) (:foreground "green4") ))
  "*The face used for directories."
  :group 'rpm-spec-faces)

(defface rpm-spec-package-face
    '(( ((class color) (background light)) (:foreground "red3") )
      ( ((class color) (background dark)) (:foreground "red3") ))
  "*The face used for files."
  :group 'rpm-spec-faces)

(defface rpm-spec-ghost-face
    '(( ((class color) (background light)) (:foreground "gray60") )
      ( ((class color) (background dark)) (:foreground "gray60") ))
  "*The face used for ghost tags."
  :group 'rpm-spec-faces)

;;; GNU emacs font-lock needs these...
(defvar rpm-spec-macro-face 'rpm-spec-macro-face "*Face for macros")
(defvar rpm-spec-tag-face 'rpm-spec-tag-face "*Face for tags")
(defvar rpm-spec-package-face 'rpm-spec-package-face "*Face for package tag")
(defvar rpm-spec-dir-face 'rpm-spec-dir-face "*Face for directory entries")
(defvar rpm-spec-doc-face 'rpm-spec-doc-face "*Face for documentation entries")
(defvar rpm-spec-ghost-face 'rpm-spec-ghost-face "*Face for %ghost files")

;;------------------------------------------------------------

(defvar rpm-no-pgp nil "Tell rpm not to sign package.")

(defvar rpm-tags-list
  '(("Autoreqprov")
    ("Buildroot")
    ("Conflicts")
    ("Copyright")
    ("%description")
    ("Distribution")
    ("Excludearch")
    ("Excludeos")
    ("Exclusivearch")
    ("Exclusiveos")
    ("%files")
    ("Group")
    ("Icon")
    ("Name")
    ("Nopatch")
    ("Nosource")
    ("%package")
    ("Packager")
    ("Patch")
    ("Prefix")
    ("Provides")
    ("Release")
    ("Requires")
    ("Serial")
    ("Source")
    ("Summary")
    ("Url")
    ("Vendor")
    ("Version"))
  "List which elements are valid tags.")

(defvar rpm-group-tags-list
  '(("Applications")
    ("Applications/Communications")
    ("Applications/Editors")
    ("Applications/Editors/Emacs")
    ("Applications/Emulators")
    ("Applications/Engineering")
    ("Applications/Spreadsheets")
    ("Applications/Databases")
    ("Applications/Graphics")
    ("Applications/Networking")
    ("Applications/Mail")
    ("Applications/Math")
    ("Applications/News")
    ("Applications/Publishing")
    ("Applications/Publishing/TeX")
    ("Applications/Publishing/SGML")
    ("Applications/Sound")
    ("Base")
    ("Base/Kernel")
    ("Utilities")
    ("Utilities/Archiving")
    ("Utilities/Console")
    ("Utilities/File")
    ("Utilities/Printing")
    ("Utilities/System")
    ("Utilities/Terminal")
    ("Utilities/Text")
    ("Daemons")
    ("Documentation")
    ("X11")
    ("X11/XFree86")
    ("X11/XFree86/Servers")
    ("X11/Applications")
    ("X11/Applications/Graphics")
    ("X11/Applications/Networking")
    ("X11/Games")
    ("X11/Games/Strategy")
    ("X11/Games/Video")
    ("X11/Amusements")
    ("X11/Utilities")
    ("X11/Libraries")
    ("X11/Window Managers")
    ("Libraries")
    ("Networking")
    ("Networking/Admin")
    ("Networking/Daemons")
    ("Networking/News")
    ("Networking/Utilities")
    ("Development")
    ("Development/Debuggers")
    ("Development/Libraries")
    ("Development/Libraries/Libc")
    ("Development/Languages")
    ("Development/Languages/Fortran")
    ("Development/Languages/Tcl")
    ("Development/Building")
    ("Development/Version Control")
    ("Development/Tools")
    ("Shells")
    ("Games"))
  "List which elements are valid group tags.")

(defvar rpm-spec-mode-syntax-table nil
  "Syntax table in use in RPM-Spec-mode buffers.")
(unless rpm-spec-mode-syntax-table
  (setq rpm-spec-mode-syntax-table (make-syntax-table))
  (modify-syntax-entry ?\\ "\\" rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?\n ">   " rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?\f ">   " rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?\# "<   " rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?/ "." rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?* "." rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?+ "." rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?- "." rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?= "." rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?% "_" rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?< "." rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?> "." rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?& "." rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?| "." rpm-spec-mode-syntax-table)
  (modify-syntax-entry ?\' "." rpm-spec-mode-syntax-table))

(defvar rpm-spec-mode-map nil
  "Keymap used in RPM Spec mode.")
(unless rpm-spec-mode-map
  (setq rpm-spec-mode-map (make-sparse-keymap))
  (and (functionp 'set-keymap-name)
       (set-keymap-name rpm-spec-mode-map 'rpm-spec-mode-map))
  (define-key rpm-spec-mode-map "\C-cp" 'rpm-build-bp)
  (define-key rpm-spec-mode-map "\C-cl" 'rpm-build-bl)
  (define-key rpm-spec-mode-map "\C-cc" 'rpm-build-bc)
  (define-key rpm-spec-mode-map "\C-ci" 'rpm-build-bi)
  (define-key rpm-spec-mode-map "\C-cb" 'rpm-build-bb)
  (define-key rpm-spec-mode-map "\C-ca" 'rpm-build-ba)
  (define-key rpm-spec-mode-map "\C-ct" 'rpm-change-timecheck)
  (define-key rpm-spec-mode-map "\C-cr" 'rpm-spec-increase-release-tag)
  (define-key rpm-spec-mode-map "\C-ce" 'rpm-add-change-log-entry)
  ;;May be better to have movement commands on \C-ck, and build on \C-c\C-k
  (define-key rpm-spec-mode-map "\C-c\C-e" 'rpm-insert-tag)
  (define-key rpm-spec-mode-map "\C-c\C-n" 'rpm-forward-section)
  (define-key rpm-spec-mode-map "\C-c\C-p" 'rpm-backward-section)
  (define-key rpm-spec-mode-map "\C-c\C-g" 'rpm-goto-section)
  (define-key rpm-spec-mode-map "\C-u\C-c\C-e" 'rpm-change-tag)
  (define-key rpm-spec-mode-map "\C-c\C-id" 'rpm-insert-description)
  (define-key rpm-spec-mode-map "\C-c\C-ic" 'rpm-insert-package)
  (define-key rpm-spec-mode-map "\C-c\C-if" 'rpm-insert-files)
  (define-key rpm-spec-mode-map "\C-c\C-is" 'rpm-insert-source)
  (define-key rpm-spec-mode-map "\C-c\C-us" 'rpm-change-source)
  (define-key rpm-spec-mode-map "\C-c\C-ip" 'rpm-insert-patch)
  (define-key rpm-spec-mode-map "\C-c\C-up" 'rpm-change-patch)
  (define-key rpm-spec-mode-map "\C-c\C-ia" 'rpm-insert-packager)
  (define-key rpm-spec-mode-map "\C-c\C-ua" 'rpm-change-packager)
  (define-key rpm-spec-mode-map "\C-c\C-iu" 'rpm-insert-url)
  (define-key rpm-spec-mode-map "\C-c\C-uu" 'rpm-change-url)
  (define-key rpm-spec-mode-map "\C-c\C-ib" 'rpm-insert-buildroot)
  (define-key rpm-spec-mode-map "\C-c\C-ub" 'rpm-change-buildroot)
  (define-key rpm-spec-mode-map "\C-c\C-ig" 'rpm-insert-group)
  (define-key rpm-spec-mode-map "\C-c\C-ug" 'rpm-change-group)
  (define-key rpm-spec-mode-map "\C-c\C-ir" 'rpm-insert-prefix)
  (define-key rpm-spec-mode-map "\C-c\C-ur" 'rpm-change-prefix)
  (define-key rpm-spec-mode-map "\C-c\C-ff" 'rpm-insert-file)
  (define-key rpm-spec-mode-map "\C-c\C-fc" 'rpm-insert-config)
  (define-key rpm-spec-mode-map "\C-c\C-fd" 'rpm-insert-doc)
  (define-key rpm-spec-mode-map "\C-c\C-fg" 'rpm-insert-ghost)
  (define-key rpm-spec-mode-map "\C-c\C-dd" 'rpm-insert-dir)
  (define-key rpm-spec-mode-map "\C-c\C-do" 'rpm-insert-docdir)
  (define-key rpm-spec-mode-map "\C-q" 'indent-spec-exp)
  ;; commented out by klakier
  ;; there is no such function
  ;;(define-key rpm-spec-mode-map "\t" 'rpm-spec-indent-command)

  ;; added by klakier
  (define-key rpm-spec-mode-map [(tab)] 'rpm-complete)
  ;; this should work with my polish xkbd layout... but it doesn't
  ;;(define-key rpm-spec-mode-map [(iso-left-tab)] 'insert-tab)
  (define-key rpm-spec-mode-map [(iso-left-tab)] (lambda () (interactive) (insert-tab)))
  (define-key rpm-spec-mode-map [(control g)]  'rpm-keyboard-quit)
  (define-key rpm-spec-mode-map [(control return)] 'rpm-expand-files1)
  (define-key rpm-spec-mode-map [(control shift return)]  'rpm-expand-files2)
  )
  
(defconst rpm-spec-mode-menu
  (purecopy '("RPM-Spec"
	      ("Insert Environment  (C-c C-e)"
	       ["Buildroot..."          rpm-insert-buildroot    t]
	       ["%description..."       rpm-insert-description  t]
	       ["%files..."             rpm-insert-files        t]
	       ["Group..."              rpm-insert-group        t]
	       ["%package..."           rpm-insert-package      t]
	       ["Packager..."           rpm-insert-packager     t]
	       ["Patch..."              rpm-insert-patch        t]
	       ["Prefix..."             rpm-insert-prefix       t]
	       ["Requires..."           rpm-insert-requires     t]
	       ["Source..."             rpm-insert-source       t]
	       ["URL..."                rpm-insert-url          t])
	      ("Change Environment  (C-u C-c C-e)"
	       ["Buildroot..."          rpm-change-buildroot    t]
	       ["Group..."              rpm-change-group        t]
	       ["Packager..."           rpm-change-packager     t]
	       ["Patch..."              rpm-change-patch        t]
	       ["Prefix..."             rpm-change-prefix       t]
	       ["Requires..."           rpm-change-requires     t]
	       ["Source..."             rpm-change-source       t]
	       ["URL..."                rpm-change-url          t])
	      ("Add file entry"
		["Regular file..."        rpm-insert-file       t]
		["Config file..."         rpm-insert-config     t]
		["Document file..."       rpm-insert-doc        t]
		["Ghost file..."          rpm-insert-ghost      t]
		"---"
		["Directory..."           rpm-insert-dir        t]
		["Document directory..."  rpm-insert-docdir     t]
		"---"
		["Default add \"%attr\" entry" (setq rpm-spec-add-attr
						    (not rpm-spec-add-attr))
		 :style toggle :selected rpm-spec-add-attr])
	      "---"
	      ["Go to section..."     rpm-mouse-goto-section t]
	      ["Forward section"      rpm-forward-section    t]
	      ["Backward sectoin"     rpm-backward-section   t]
	      "---"
	      ["Execute \"%prep\" stage"   rpm-build-bp      t]
	      ["Do a \"list check\""       rpm-build-bl      t]
	      ["Do the \"%build\" stage"   rpm-build-bc      t]
	      ["Do the \"%install\" stage" rpm-build-bi      t]
	      ["Build binary package"      rpm-build-bb      t]
	      ["Build binary and source"   rpm-build-ba      t]
	      "---"
	      ["Add changelog entry..."    rpm-add-change-log-entry      t]
	      ["Change timecheck value..." rpm-change-timecheck          t]
	      ["Increase release-tag"      rpm-spec-increase-release-tag t]
	      "---"
	      ["Short Circuit" (setq rpm-spec-short-circuit
				     (not rpm-spec-short-circuit))
	       :style toggle :selected rpm-spec-short-circuit]
	      ["Clean" (setq rpm-spec-clean (not rpm-spec-clean))
	       :style toggle :selected rpm-spec-clean]
	      ["Testing only" (setq rpm-spec-test (not rpm-spec-test))
	       :style toggle :selected rpm-spec-test]
	      ["PGP Sign" (setq rpm-spec-sign-pgp (not rpm-spec-sign-pgp))
	       :style toggle :selected rpm-spec-sign-pgp]
	      "---"
	      ["About rpm-spec-mode"       rpm-about-rpm-spec-mode        t]
	      )))

(defvar rpm-spec-font-lock-keywords
  '(
    ("%[a-zA-Z-]+" 0 rpm-spec-macro-face)
    ("^\\([a-zA-Z0-9]+\\):" 1 rpm-spec-tag-face)
    ("%\\(define\\|files\\|package\\|description\\)[ \t]+\\([^ \t\n-]+\\)"
     (2 rpm-spec-package-face))
    ("%dir[ \t]+\\([^ \t\n]+\\)[ \t]*" 1 rpm-spec-dir-face)
    ("%doc\\(\\|dir\\)[ \t]+\\(.*\\)\n" 2 rpm-spec-doc-face)
    ("%\\(ghost\\|config\\)[ \t]+\\(.*\\)\n" 2 rpm-spec-ghost-face)
    ("^%.+-[a-zA-Z][ \t]+\\([a-zA-Z0-9\.-]+\\)" 1 rpm-spec-doc-face)
    ("^\\(.+\\)(\\([a-zA-Z]\\{2,2\\}\\)):" 
     (1 rpm-spec-tag-face)
     (2 rpm-spec-doc-face))
    ("^\\*\\(.\\{16,16\\}\\).\\(.*\\)\\(<.*>\\)\n"
     (1 rpm-spec-dir-face)
     (2 rpm-spec-package-face)
     (3 rpm-spec-tag-face))
    ("%{[a-zA-Z_-]+}" 0 rpm-spec-doc-face)
    )
  "Additional expressions to highlight in RPM Spec mode.")

;;Initialize font lock for xemacs
(put 'rpm-spec-mode 'font-lock-defaults '(rpm-spec-font-lock-keywords))

(defvar rpm-spec-mode-abbrev-table nil
  "Abbrev table in use in RPM-Spec-mode buffers.")
(define-abbrev-table 'rpm-spec-mode-abbrev-table ())

;;------------------------------------------------------------

;;;###autoload
(defun rpm-spec-mode ()
  "Major mode for editing spec files.
This is much like C mode except for the syntax of comments.  It uses
the same keymap as C mode and has the same variables for customizing
indentation.  It has its own abbrev table and its own syntax table.

Turning on RPM Spec mode calls the value of the variable `rpm-spec-mode-hook'
with no args, if that value is non-nil."
  (interactive)
  (kill-all-local-variables)
  (require 'cc-mode)
  (use-local-map rpm-spec-mode-map)
  (setq major-mode 'rpm-spec-mode)
  (setq mode-name "RPM-SPEC")
  (setq rpm-buffer-name "*RPM Background*")
  (setq local-abbrev-table rpm-spec-mode-abbrev-table)
  (set-syntax-table rpm-spec-mode-syntax-table)

  (require 'easymenu)
  (easy-menu-define rpm-spec-call-menu rpm-spec-mode-map 
		    "Post menu for rpm-spec-mode" rpm-spec-mode-menu)
  (easy-menu-add rpm-spec-mode-menu)

  (if (= (buffer-size) 0)
      (rpm-spec-initialize))

  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
;  (make-local-variable 'indent-line-function)
;  (setq indent-line-function 'c-indent-line)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'comment-start)
  (setq comment-start "# ")
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (make-local-variable 'comment-column)
  (setq comment-column 32)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "#+ *")
;  (make-local-variable 'comment-indent-function)
;  (setq comment-indent-function 'c-comment-indent)
  ;;Initialize font lock for GNU emacs.
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(rpm-spec-font-lock-keywords nil t))
  (run-hooks 'rpm-spec-mode-hook))

(defun rpm-command-filter (process string)
  "Filter to process normal output."
  (save-excursion
    (set-buffer (process-buffer process))
    (save-excursion
      (goto-char (process-mark process))
      (insert-before-markers string)
      (set-marker (process-mark process) (point)))))

;;------------------------------------------------------------

(defun rpm-add-change-log-entry (&optional change-log-entry)
  "Find change log and add an entry for today."
  (interactive "p")
  (save-excursion
    (goto-char (point-min))
    (if (search-forward-regexp "^%changelog[ \t]*$" nil t)
	(let ((string (concat "* " (substring (current-time-string) 0 11)
			      (substring (current-time-string) -4) " "
			      (user-full-name) " <" user-mail-address ">")))
	  (if (not (search-forward string nil t))
	      (insert "\n" string "\n\n")
	    (next-line 2)
	    (beginning-of-line))
	  (if (eq change-log-entry 1)
	      (insert "- " (read-from-minibuffer "Changelog entry: ") "\n")
	    (insert "- " change-log-entry)))
      (message "No \"%%changelog\" entry found..."))))

;;------------------------------------------------------------

(defun rpm-insert-f (&optional filetype filename)
  "Insert new %files entry."
  (save-excursion
    (goto-char (point-min))
    (if (search-forward-regexp "^%files[ ]*$" nil t)
	(progn
	  (if (or (eq filename 1) (not filename))
	      (insert "\n" (read-file-name
			    (concat filetype "filename: ") "" "" nil))
	    (insert "\n" filename))
	  (beginning-of-line)
	  (if rpm-spec-add-attr (insert "%attr(-, root, root) "))
	  (insert filetype))
      (message "No \"%%files\" entry found..."))))
    

(defun rpm-insert-file (&optional filename)
  "Insert regular file."
  (interactive "p")
  (rpm-insert-f "" filename))
  
(defun rpm-insert-config (&optional filename)
  "Insert config file."
  (interactive "p")
  (rpm-insert-f "%config " filename))

(defun rpm-insert-doc (&optional filename)
  "Insert doc file."
  (interactive "p")
  (rpm-insert-f "%doc " filename))

(defun rpm-insert-ghost (&optional filename)
  "Insert ghost file."
  (interactive "p")
  (rpm-insert-f "%ghost " filename))

(defun rpm-insert-dir (&optional dirname)
  "Insert directory."
  (interactive "p")
  (rpm-insert-f "%dir " dirname))

(defun rpm-insert-docdir (&optional dirname)
  "Insert doc directory."
  (interactive "p")
  (rpm-insert-f "%docdir " dirname))

;;------------------------------------------------------------

(defun rpm-insert (&optional what file-completion)
  "Insert given tag.  Use file-completion if argument is t."
  (beginning-of-line)
  (if (not what)
      (setq what (completing-read "Tag: " rpm-tags-list)))
  (if (string-match "^%" what)
      (setq read-text (concat "Packagename for " what ": ")
	    insert-text (concat what " "))
    (setq read-text (concat what ": ")
	  insert-text (concat what ": ")))
  (cond
   ((string-equal what "Group")
    (rpm-insert-group))
   ((string-equal what "Source")
    (rpm-insert-n "Source"))
   ((string-equal what "Patch")
    (rpm-insert-n "Patch"))
   (t
    (if file-completion
	(insert insert-text (read-file-name (concat read-text) "" "" nil) "\n")
      (insert insert-text (read-from-minibuffer (concat read-text)) "\n")))))

(defun rpm-insert-n (what &optional arg)
  "Insert given tag with possible number."
  (save-excursion
    (goto-char (point-max))
    (if (search-backward-regexp (concat "^" what "\\([0-9]*\\):") nil t)
	(let ((release (1+ (string-to-int (match-string 1)))))
	  (next-line 1)
	  (beginning-of-line)
	  (let ((default-directory "/usr/src/redhat/SOURCES/"))
	    (insert what (int-to-string release) ": "
		    (read-file-name (concat what "file: ") "" "" nil) "\n")))
      (goto-char (point-min))
      (rpm-end-of-section)
      (insert what ": " (read-from-minibuffer (concat what "file: ")) "\n"))))

(defun rpm-change (&optional what arg)
  "Update given tag."
  (save-excursion
    (if (not what)
	(setq what (completing-read "Tag: " rpm-tags-list)))
    (cond
     ((string-equal what "Group")
      (rpm-change-group))
     ((string-equal what "Source")
      (rpm-change-n "Source"))
     ((string-equal what "Patch")
      (rpm-change-n "Patch"))
     (t
      (goto-char (point-min))
      (if (search-forward-regexp (concat "^" what ":[ ]*\\(.*\\)$") nil t)
	  (replace-match
	   (concat what ": " (read-from-minibuffer
			      (concat "New " what ": ") (match-string 1))))
	(message (concat what " tag not found...")))))))

(defun rpm-change-n (what &optional arg)
  "Change given tag, possible numbers."
  (save-excursion
    (goto-char (point-min))
    (let ((number (read-from-minibuffer (concat what " number: "))))
      (if (search-forward-regexp
	   (concat "^" what number ":[ ]*\\(.*\\)") nil t)
	  (let ((default-directory "/usr/src/redhat/SOURCES/"))
	    (replace-match
	     (concat what number ": "
		     (read-file-name (concat "New " what number " file: ")
					 "" "" nil (match-string 1)))))
	(message (concat what " number \"" number "\" not found..."))))))

(defun rpm-insert-group (&optional arg)
  "Insert Group tag."
  (interactive "p")
  (beginning-of-line)
  (insert "Group: " (completing-read "Group: " rpm-group-tags-list) "\n"))

(defun rpm-change-group (&optional arg)
  "Update Group tag."
  (interactive "p")
  (save-excursion
    (goto-char (point-min))
    (if (search-forward-regexp "^Group: \\(.*\\)$" nil t)
	(replace-match
	 (concat "Group: "
		 (insert (completing-read "Group: " rpm-group-tags-list
					  nil nil (match-string 1)))))
      (message "Group tag not found..."))))

(defun rpm-insert-tag (&optional arg)
  "Insert a tag."
  (interactive "p")
  (rpm-insert))

(defun rpm-change-tag (&optional arg)
  "Change a tag."
  (interactive "p")
  (rpm-change))

(defun rpm-insert-source (&optional arg)
  "Insert Source tag."
  (interactive "p")
  (rpm-insert-n "Source"))

(defun rpm-change-source (&optional arg)
  "Change Source tag."
  (interactive "p")
  (rpm-change-n "Source"))

(defun rpm-insert-patch (&optional arg)
  "Insert Patch tag."
  (interactive "p")
  (rpm-insert-n "Patch"))

(defun rpm-change-patch (&optional arg)
  "Change Source tag."
  (interactive "p")
  (rpm-change-n "Patch"))

(defun rpm-insert-url (&optional arg)
  "Insert URL tag."
  (interactive "p")
  (rpm-insert "Url"))

(defun rpm-change-url (&optional arg)
  "Update URL tag."
  (interactive "p")
  (rpm-change "Url"))

(defun rpm-insert-buildroot (&optional arg)
  "Insert Buildroot tag."
  (interactive "p")
  (rpm-insert "Buildroot" t))

(defun rpm-change-buildroot (&optional arg)
  "Update Buildroot tag."
  (interactive "p")
  (rpm-change "Buildroot"))

(defun rpm-insert-requires (&optional arg)
  "Insert Requires tag."
  (interactive "p")
  (rpm-insert "Requires" t))

(defun rpm-change-requires (&optional arg)
  "Update Requires tag."
  (interactive "p")
  (rpm-change "Requires"))

(defun rpm-insert-summary (&optional arg)
  "Insert Summary tag."
  (interactive "p")
  (rpm-insert "Summary"))

(defun rpm-change-summary (&optional arg)
  "Update Summary tag."
  (interactive "p")
  (rpm-change "Summary"))

(defun rpm-insert-description (&optional arg)
  "Insert %description tag."
  (interactive "p")
  (rpm-insert "%description"))

(defun rpm-insert-files (&optional arg)
  "Insert %files tag."
  (interactive "p")
  (rpm-insert "%files"))

(defun rpm-insert-package (&optional arg)
  "Insert %package tag."
  (interactive "p")
  (rpm-insert "%package"))

(defun rpm-insert-packager (&optional arg)
  "Insert Packager tag."
  (interactive "p")
  (beginning-of-line)
  (insert "Packager: " (user-full-name) " <" user-mail-address ">\n"))

(defun rpm-change-packager (&optional arg)
  "Update Packager tag."
  (interactive "p")
  (rpm-change "Packager"))

;;------------------------------------------------------------

(defun rpm-current-section nil
  (interactive)
  (save-excursion
    (rpm-forward-section)
    (rpm-backward-section)
    (let ((sect (buffer-substring (match-beginning 1) (match-end 1))))
      ;; if this is "package" or beginning od buffer, return "preamble" /klakier
      (if (or (bobp) (string= sect "package")) "preamble" sect))))

(defun rpm-backward-section nil
  "Move backward to the beginning of the previous section.
Go to beginning of previous section."
  (interactive)
  (or (re-search-backward rpm-section-regexp nil t)
      (goto-char (point-min))))

(defun rpm-beginning-of-section nil
  "Move backward to the beginning of the current section.
Go to beginning of current section."
  (interactive)
  (or (and (looking-at rpm-section-regexp) (point))
      (re-search-backward rpm-section-regexp nil t)
      (goto-char (point-min))))

(defun rpm-forward-section nil
  "Move forward to the beginning of the next section."
  (interactive)
  (forward-char)
  (if (re-search-forward rpm-section-regexp nil t)
      (progn (forward-line 0) (point))
    (goto-char (point-max))))

(defun rpm-end-of-section nil
  "Move forward to the end of this section."
  (interactive)
  (forward-char)
  (if (re-search-forward rpm-section-regexp nil t)
      (forward-line -1)
    (goto-char (point-max)))
  (while (looking-at paragraph-separate)
    (forward-line -1))
  (forward-line 1)
  (point))

(defun rpm-goto-section (section)
  "Move point to the beginning of the specified section; 
leave point at previous location."
  (interactive "ssection:")
  (push-mark)
  (goto-char (point-min))
  (or 
   (equal section "preamble")
   (re-search-forward (concat "^%" section "\\b") nil t)
   (let ((s (cdr rpm-sections)))
     (while (not (equal section (car s)))
       (re-search-forward (concat "^%" (car s) "\\b") nil t)
       (setq s (cdr s)))
     (if (re-search-forward rpm-section-regexp nil t)
	 (forward-line -1) (goto-char (point-max)))
     (insert "\n%" section "\n"))))

(defun rpm-mouse-goto-section (&optional section)
  (interactive 
   (x-popup-menu
    last-nonmenu-event 
    (list "sections" 
	  (cons "sections" (mapcar (lambda (e) (list e e)) rpm-sections))
	  (cons "scripts" (mapcar (lambda (e) (list e e)) rpm-scripts))
	  )))
  (and section ;if user doesn't pick a section, exit quietly.
       (if (member section rpm-sections)
	   (rpm-goto-section section)
	 (goto-char (point-min))
	 (or (re-search-forward (concat "^%" section "\\b") nil t)
	     (and (re-search-forward "^%files\\b" nil t) (forward-line -1))
	     (goto-char (point-max))))))


;;------------------------------------------------------------

(defun rpm-build (buildoptions)
  "Build this rpm-package."
  (rpm-process-check rpm-buffer-name)

  (if (get-buffer rpm-buffer-name)
      (kill-buffer rpm-buffer-name))
  (create-file-buffer rpm-buffer-name)
  (display-buffer rpm-buffer-name)
  (setq buildoptions (list buildoptions buffer-file-name))
  (if (or rpm-spec-short-circuit rpm-spec-test)
      (setq rpm-no-pgp t))
  (if rpm-spec-clean
      (setq buildoptions (cons "--clean" buildoptions)))
  (if rpm-spec-short-circuit
      (setq buildoptions (cons "--short-circuit" buildoptions)))
  (if rpm-spec-test
      (setq buildoptions (cons "--test" buildoptions)))
  (if (and rpm-spec-sign-pgp (not rpm-no-pgp))
      (setq buildoptions (cons "--sign" buildoptions)))
  (goto-char (point-max) (get-buffer rpm-buffer-name))
  (let ((process
         (apply 'start-process "rpm" rpm-buffer-name "rpm" buildoptions)))
    (if (and rpm-spec-sign-pgp (not rpm-no-pgp))
	(let ((rpm-passwd-cache (read-passwd "PGP passphrase: ")))
	  (process-send-string process (concat rpm-passwd-cache "\n"))))
    (set-process-filter process 'rpm-command-filter)))

(defun rpm-build-bp (&optional arg)
  "Run a `rpm -bp'."
  (interactive "p")
  (if rpm-spec-short-circuit
      (message "Cannot run `rpm -bp' with short-circuit")
    (setq rpm-no-pgp t)
    (rpm-build "-bp")))

(defun rpm-build-bl (&optional arg)
  "Run a `rpm -bl'."
  (interactive "p")
  (if rpm-spec-short-circuit
      (message "Cannot run `rpm -bl' with short-circuit")
    (setq rpm-no-pgp t)
    (rpm-build "-bl")))

(defun rpm-build-bc (&optional arg)
  "Run a `rpm -bc'."
  (interactive "p")
  (setq rpm-no-pgp t)
  (rpm-build "-bc"))

(defun rpm-build-bi (&optional arg)
  "Run a `rpm -bi'."
  (interactive "p")
  (setq rpm-no-pgp t)
  (rpm-build "-bi"))

(defun rpm-build-bb (&optional arg)
  "Run a `rpm -ba'."
  (interactive "p")
  (if rpm-spec-short-circuit
      (message "Cannot run `rpm -bb' with short-circuit")
    (setq rpm-no-pgp nil)
    (rpm-build "-bb")))

(defun rpm-build-ba (&optional arg)
  "Run a `rpm -ba'."
  (interactive "p")
  (if rpm-spec-short-circuit
      (message "Cannot run `rpm -ba' with short-circuit")
    (setq rpm-no-pgp nil)
    (rpm-build "-ba")))

(defun rpm-process-check (buffer)
  "Check if BUFFER has a running process.
If so, give the user the choice of aborting the process or the current
command."
  (let ((process (get-buffer-process (get-buffer buffer))))
    (if (and process (eq (process-status process) 'run))
	(if (yes-or-no-p (concat "Process `" (process-name process)
				 "' running.  Kill it? "))
	    (delete-process process)
	  (error "Cannot run two simultaneous processes ...")))))

;;------------------------------------------------------------

(defun rpm-change-timecheck (&optional arg)
  "Change the value for timecheck."
  (interactive "p")
  (setq rpm-spec-timecheck
	(read-from-minibuffer "New timecheck: " rpm-spec-timecheck)))

(defun rpm-spec-increase-release-tag (&optional arg)
  "Increase the release tag by 1."
  (interactive "p")
  (save-excursion
    (goto-char (point-min))
    (if (search-forward-regexp "^Release: \\([0-9]+\\)\\(.*\\)" nil t)
	(let ((release (1+ (string-to-int (match-string 1)))))
	  (setq release (concat (int-to-string release) (match-string 2)))
	  (replace-match (concat "Release: " release))
	  (message (concat "Release tag changed to " release ".")))
      (message "No Release tag found..."))))

;;------------------------------------------------------------

(defun rpm-spec-initialize ()
  "Create a default spec file if one does not exist or is empty."
  (let (file name version (release "1"))
    (setq file (if (buffer-file-name)
		   (file-name-nondirectory (buffer-file-name))
		 (buffer-name)))
    ;; try to find version number and release from the filename
    (cond
     ((eq (string-match "\\([^-]*\\)-\\([^-]*\\)-\\(.*\\).spec" file) 0)
      (setq name (match-string 1 file))
      (setq version (match-string 2 file))
      (setq release (match-string 3 file)))
     
     ((eq (string-match "\\(.*\\)-\\([^-]*\\).spec" file) 0)
      (setq name (match-string 1 file))
      (setq version (match-string 2 file)))

     ((eq (string-match "\\(.*\\).spec" file) 0)
      (setq name (match-string 1 file))))
    
    (insert "# " file
	    "\n\nSummary: "
	    "\nName: " (or name "")
	    "\nVersion: " (or version "") 
	    "\nRelease: " (or release "")
	    "\nCopyright: \nGroup: \n\n%description\n"
	    (if rpm-initialize-sections
		"\n%changelog\n%prep\n%setup\n\n%build\n\n%install\n\n%files\n"
	      "")
	    "\n# end of file\n")
    (rpm-add-change-log-entry "initialization of spec file.\n")))

;;------------------------------------------------------------

(defun rpm-about-rpm-spec-mode (&optional arg)
  "About rpm-spec-mode."
  (interactive "p")
  (message "Made by Stig Bjrlykke, <stigb@tihlde.hist.no>"))

(provide 'rpm-spec-mode)



;; !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;; !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;; !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
;; my stuff starts here /klakier

;; Q rpm: can macro A be used before definition?
;; Q rpm: can macros be redefined? sure...
;; hmm, matching is case insensitive by default... what about matching RPM_BUILD_ROOT

;; BUGS:
;; - rpm-current-section does not recognize preun (only this section, why?)
;; TODO: 
;; - %config(option) is not font-locked
;; - font-lock for CVS changelog

;;test only
;;(global-set-key [(insert)] 'rpm-complete)


(defun rpm-expand-files1 ()
  (interactive "*")
  (rpm-expand-files nil)
)
(defun rpm-expand-files2 ()
  (interactive "*")
  (rpm-expand-files t)
)

(defun rpm-expand-files (line) 
  "Expand file pattern. If line is t expands in separate lines, copying beginning of line" 
  "If line is nil expands in one line" 
  ;;(interactive "*")
  (let ((sect (rpm-current-section))
	(string (string-before))
	(pat nil)
	(dir nil)
	(pre nil)
	(post nil)
	)
    (setq pat (file-name-nondirectory string))
    (setq dir (or (file-name-directory string) ""))
    ;; set the text that will be reproduced before each expanded filename
    (setq pre
	  (if (not line) dir ;; either dir
	    ;; or part of line to the beginning of pat
	    (save-excursion 
	      (let ((p (point)))
		;; first save tail of line
		(end-of-line)
		(setq post (buffer-substring p (point)))
		;; then return head of line
		(beginning-of-line)
		(buffer-substring (point) (- p (length pat)))
		))))
    (when (setq dir (rpm-parse dir))
      (if (cond ((string= (rpm-current-section) "files")
		 (setq dir (if (string-match "^%doc" (first-string)) 
			       (concat (rpm-dir-src) dir)
			     (concat (rpm-dir-buildroot) dir)
			     )))
		((string= (rpm-current-section) "install")
		 (setq dir (concat (rpm-dir-src) dir)
			     ))
		) 
	  ;; end of 'if' condition
	  ;; begin of 'true' form
	  (progn (if line
		     ;; insert comment
		     (progn (beginning-of-line) 
			    (insert "## BEGIN of expanded line: ") 
			    (end-of-line)
			    (insert "\n")
			    )
		   ;; or delete expanded string
		   (progn  (backward-delete-char (length string))))
		 ;; execute pipe with output to buffer
		 (call-process "sh" nil t nil "-c" 
			       (concat "cd " dir " && ls -d1 " pat 
				       " | awk -v ORS=\"" (if line "\n" " ") "\""
				       " '{print \"" pre "\" $0 \"" post "\"}'"))
		 (if line 
		     ;; add comment
		     (progn (insert (concat "## END of expanded line: " pre pat post)))
		   ))
		 
	;; begin of 'false' form
	(message "Expanding not allowed in this section."))
      )))



(defun rpm-complete
  "Complete tag, filename"
  (interactive "*")
  (let ((tab nil) ;; completion table
	(pattern nil)
	(fstring (first-string)) ; first string in line
	(string (string-before)) ; string before point
	;;(c (char-before))
	(sect (rpm-current-section))
	(file nil)
	(dir nil)
	(tmp nil)
	;;(here (point))
	)
    (setq pattern string) ;; first letters to complete


    (cond ((setq tmp (string-match "%{[^}]*$" string)) ;; "%{[^}]*$"
	   ;; general macro completion rules (%{version}, and so on...)
	   (setq pattern (substring string tmp))
	   (setq tab (make-completion-table (append rpm-macros-completion-list1
						    rpm-macros-completion-list2)))
	   )
	  ((string-match "%$" string)
	   (setq pattern "%")
	   (setq tab (make-completion-table rpm-macros-completion-list2)) ; '%' is completed to %{_ macros only
	   )
	  ((string= sect "preamble")
	   (cond ((string= fstring string) ;; this is first string in line - maybe tag?
		  (setq tab (make-completion-table rpm-preamble-completion-list))
		  )
		 ;; want to complete filename from SOURCES dir?
		 ((or (string-match "^Source[0-9]*:"  fstring)
		      (string-match "^Patch[0-9]*:"  fstring))
		  (if (string= "" (file-name-nondirectory string))
		      (if (string-match "^Source[0-9]*:"  fstring)
			  (insert "%{name}-%{version}") (insert "%{name}-"))
		    ;; else complete filename
		    (progn (setq file (file-name-nondirectory string))
			   (setq dir rpm-dir-sources))))
		 ;; group completions
		 ((string-match "^Group:"  fstring)
		  (setq tab (make-completion-table rpm-groups-completion-list)))
		 ((string-match "^%define"  fstring)
		  ;; enable inserting ordinary tabs
		  (insert-tab))
		 ))
	  ((string= "description" sect)
	   (cond ((string= "" string) ;; we are at the beginning of line
		  (insert-tab))
		 ((string= fstring string) ;; if first word in line: tag
		  (setq tab (make-completion-table rpm-description-completion-list)))
		 (t  ;; else
		  (insert-tab))
		 ))
	  ((string= "prep" sect)
	   (setq tab (make-completion-table rpm-prep-completion-list))
	   )
	  ((string= "build" sect)
	   (cond ((string= string "") ;; 
		  (insert-tab))
		 ((and (string= fstring string) ;; complete command or tag
		       (not (string-match "^\\." string)) ; beginning of ./command ?
		       (not (string-match "^/"  string))) ; begining of /path/command ?
		  (setq tab (make-completion-table rpm-build-completion-list)))
		 ((setq tmp (string-match "\$[A-Z_]*$" string)) ;; '$' complete to '$RPM_OPT_FLAGS'
		  (setq pattern (substring string tmp))
		  (setq tab (make-completion-table '("$RPM_OPT_FLAGS"))))
		 (t ;; treat as filename
		  (setq file (replace-in-string string ".*=" "")) ; --option=/path
		  (setq dir (rpm-dir-src)))
		 ))	   
	  ((string= "install" sect)
	   (cond ((string= string "") ;; 
		  (insert-tab))
	         ((and (string= fstring string) ; first string in line?
		       (not (string-match "^\\." string)) ; beginning of ./command ?
		       (not (string-match "^/"  string))) ; begining of /path/command ?
		  ;; complete tag or common command from list
		  (setq tab (make-completion-table rpm-install-completion-list)))
		 ((setq tmp (string-match "\$[A-Z_]*$" string)) ;; '$' complete to '$RPM_BUILD_ROOT'
		  (setq pattern (substring string tmp))
		  (setq tab (make-completion-table '("$RPM_BUILD_ROOT"))))
		 (t  ;; else treat as filename
		  (setq file (replace-in-string string ".*=" ""))
		  (setq dir (rpm-dir-src)))
		 ))	   
	  ((or (string-match "^pre" sect)
	       (string-match "^post" sect)
	       (string= sect "clean"))
	   ;; complete entire line not only the last string
	   (setq pattern (current-line))
	   (setq tab (make-completion-table rpm-scripts-completion-list))
	   )
	  ((string= sect "files") ;; %files section is treated specially: macro completions differ
	   (cond ((string-match "^%[^{]" string)  ;; (elt string 0) ?%)
		  ;; complete files-specific macro like %doc, %config
		  (setq tab (make-completion-table rpm-files-completion-list)))
		 (t 		 ;; else filename completion
		  (if (string-match "^%doc" fstring)
		      (progn (setq file string)
			     (setq dir (rpm-dir-src))) ;; in %doc we take files from sources
		    (if (string= "" string) ;; on empty string complete path macros ('%{_')
			(setq tab (make-completion-table rpm-macros-completion-list2))
		      (progn (setq file (concat (rpm-dir-buildroot) string))) ;; take files from bulidroot
		      )))
		 ))
	  
	  );; section cond
    
	  
    (when (setq file (rpm-parse file))
      ;; if 'file' variable is set by some of preceding rules, 
      ;; prepare filename completion table for this file
      (unless (string-match "^/" file) (setq file (concat dir file)))
      (setq tab
	    (make-completion-table
	     (or
	      ;; FIXME: why sort doesn't work
	      (sort (file-name-all-completions (setq pattern (file-name-nondirectory file))
					       (file-name-directory file)) 'string<) ; sort criterium
	      ;; if dir not exists, file-name-all completions returns null. 
	      ;; I change it to list with empty string cause I want it to be processed 
	      ;; to produce error message
	      (list "") 
	      )))
      )
    
    (when tab
      (let ((completion (try-completion pattern tab)))
	(cond ((null completion)
	       ;;(goto-char here)
	       (message "Can't find completion for \"%s\"" pattern)
	       (ding))
	      ((eq completion t)
	       ;;(goto-char here)
	       (message "Complete!!!")
	       )
	      ((not (string= pattern completion))
	       ;(delete-backward-char (length pattern))
	       ;(insert completion)
	       (insert (substring completion (length pattern)))
	       (message "Completion not finished")
	       (setq completion (try-completion completion tab))
	       ;; is complete already
	       (if (eq completion t)
		   (progn (if (get-buffer "*RPM-Completions*")
			      (progn (kill-buffer "*RPM-Completions*")
				     (if rpm-window-configuration 
					 (set-window-configuration rpm-window-configuration)
					 (setq rpm-window-configuration nil))))
			  (message "Completion finished"))))
	      (t
	       ;;(goto-char here)
	       (message "Making completion list...")
	       (let ((list (all-completions pattern tab)))
		 (unless rpm-window-configuration (setq rpm-window-configuration (current-window-configuration)))
		 (with-output-to-temp-buffer "*RPM-Completions*"
		   (display-completion-list list))
		 (message "Making completion list...%s" "done"))
	       )
	      )
	)
      )
    )
  )


(defun rpm-adapterize ()
  "Process spec file with PLD 'adapter'. Saves buffer before processing"
  (interactive "*")
  (if rpm-dir-specs
      (let ((adapter (concat rpm-dir-specs "adapter.awk")))
	(if (file-executable-p adapter)
	    (progn
	      (save-buffer)
	      (erase-buffer)
	      (message "Adapterizing...")
	      (call-process adapter nil t nil (buffer-file-name))
	      ;;(insert "Adapterization failed.\nUse 'C-/' or 'C-x u' to undo."))
	      (beginning-of-buffer))
	  (message (concat adapter " nonexistent or not executable!"))
	  ))))

(defun rpm-patch-create ()
  "Copy directory with unpacked sources to the directory with .new extension"
  (interactive "*")
  (let ((dir (directory-file-name (rpm-dir-src)))
	(ndir nil)
	(create-dir nil)
	(command nil)
	(patch nil)
	(buf nil))
    (setq ndir (concat dir ".new"))
    (cond ((not (file-exists-p dir)) 
	   (when (y-or-n-p (concat dir " does not exist. Execute %prep?")) (rpm-build-bp) ))
	  ((not (file-exists-p ndir)) (setq create-dir "-a")) ;; no ndir - just create one
	  ((not (y-or-n-p (concat "Create patch based on existing " 
				  (file-name-nondirectory ndir) " directory ?")))
	   ;; answer was "no" - ask to recreate ndir
	   (when (yes-or-no-p (concat "Recreate " ndir " ?")) 
	     (setq create-dir "-af")
	     (if (string-match "rpm/BUILD/.+" ndir) ;; just a hack to ensure the path is correct before deleting
		 (call-process "rm" nil t nil "-fr" ndir)
	       (error (concat "Error: " ndir " looks to be incorrect!!!")))
	     ))
	  ((and  ;; otherwise try to create patch
	    (setq patch (read-file-name 
			 "Patch name: " rpm-dir-sources
			 nil nil (concat (rpm-find-value "name") "-REASON.patch")))
	    (setq command (concat  "diff -Nur --exclude '*~' " 
				   (file-name-nondirectory dir) " " 
				   (file-name-nondirectory ndir)))
					; >" rpm-dir-sources patch))
	    ;; open new file-buffer for patch
	    (setq buf (create-file-buffer (concat rpm-dir-sources patch)))
	    (set-buffer buf)
	    (switch-to-buffer buf)
	    (setq default-directory rpm-dir-sources)
	    (or (set-visited-file-name patch) t) ;always t
	    (message (concat "Running " command))
	    ;; put the diff output into new buffer
	    (call-process "sh" nil buf nil "-c" (concat "cd " rpm-dir-build " && " command))
	    ))
	  )
    (when create-dir
      (message (concat "Creating " ndir))
      (call-process "cp" nil t nil create-dir dir ndir)
      (message "Edit sources, then use rpm-patch-create again to create patch-file")
)))


(defun rpm-add-package ()
  "Inserts new package preamble, new empty %description, and empty %files section"
  ;; put the point before prep 
  ;;TODO: do it in more elegant way
  (interactive "*")
    (let ((pkg (read-string "Package name: " "devel")))
      ;; first add %files section
      (rpm-goto-section "changelog")
      (previous-line 1) ;; there is usually %define date before %changelog
      (beginning-of-line)
      (insert "%files ") (insert pkg)
      (insert "\n%defattr(644,root,root,755)\n")
      (cond ((string= pkg "devel")
	     (insert "%{_includedir}/*\n")
	     (insert "%{_libdir}/*.so\n")
	     ))
      (insert "\n")
      ;; then add %package and %description section
      (rpm-goto-section "prep")
      (beginning-of-line)
      (cond ((string= pkg "devel")
	     (insert "%package devel\n")
	     (insert "Summary:	%{name} header files\n")
	     (insert "Summary(pl):	Pliki nagwkowe %{name}\n")
	     (insert "Group:		Development/Libraries\n")
	     (insert "Group(fr):	Development/Librairies\n")
	     (insert "Group(pl):	Programowanie/Biblioteki\n")
	     (insert "Requires:	%{name} = %{version}\n\n")
	     (insert "%description devel\n")
	     (insert "%{name} header files.\n\n")
	     (insert "%description -l pl devel\n")
	     (insert "Pliki nagwkowe %{name}.\n\n")
	     )
	    ((string= pkg "static")
	     (insert "%package static\n")
	     (insert "Summary:	Static %{name} libraries\n")
	     (insert "Summary(pl):	Biblioteki statyczne %{name}\n")
	     (insert "Group:		Development/Libraries\n")
	     (insert "Group(fr):	Development/Librairies\n")
	     (insert "Group(pl):	Programowanie/Biblioteki\n")
	     (insert "Requires:	%{name}-devel = %{version}\n\n")
	     (insert "%description static\n")
	     (insert "Static %{name} libraries.\n\n")
	     (insert "%description -l pl static\n")
	     (insert "Biblioteki statyczne %{name}.\n\n")
	     )
	    ((string= pkg "utils")
	     (insert "%package utils\n")
	     (insert "Summary:	%{name} utilities\n")
	     (insert "Summary(pl):	Narzdzia %{name}\n")
	     (insert "Group:		-\n")
	     (insert "Requires:	%{name} = %{version}\n\n")
	     (insert "%description utils\n")
	     (insert "%{name} utilities.\n\n")
	     (insert "%description -l pl utils\n")
	     (insert "Narzdzia zwizane z %{name}.\n\n")
	     )
	    (t;; default
	     (insert "%package ") (insert pkg) (insert "\n")
	     (insert "Summary:	%{name} header files\n")
	     (insert "Summary(pl):	Pliki nagwkowe %{name}\n")
	     (insert "Group:		-\n")
	     (insert "Requires:	%{name} = %{version}\n\n")
	     (insert "%description ") (insert pkg) (insert "\n\n")
	     (insert "%description -l pl ") (insert pkg) (insert "\n\n")
	     )
	    )
      (previous-line 1)
      ))

;;(defun rpm-cvs-commit ()
;;)

(defun rpm-keyboard-quit () "Delete *RPM-Competions* buffer and restores windows layout. Bind locally to C-g"
  (interactive
   (if rpm-window-configuration
       (progn (set-window-configuration rpm-window-configuration)
	      (setq rpm-window-configuration nil))
     (keyboard-quit)
     )))

(setq rpm-window-configuration nil)


;; ###################### parsing functions ##########################

(defun rpm-eval (str) "Parse str with 'rpm --eval'"
  (if (or (null str) (string= "" str)) 
      nil
    (save-excursion
      (let ((p (point)))
	(call-process "rpm" nil t nil "--eval" str)
	(prog1
	    (buffer-substring p (1- (point))) ; don't want trailing newline
	  (delete-backward-char (- (point) p)))))))


(defun rpm-parse (str) 
  "Parse str to evaluate macros"
  (when str
    ;; first substitute RPM_BUILD_ROOT if present
    (let ((beg (string-match "\$RPM_BUILD_ROOT" str)))
      (if beg (setq str (concat (substring str 0 beg) 
				(directory-file-name (rpm-dir-buildroot))
				(substring str (+ beg 15))))))
    ;; then try to parse without rpm's help
    ;; if it failed, pass to 'rpm --eval' and try again
    (or (rpm-parse-rec str)
	(rpm-parse-rec (rpm-eval str)))))

(defun rpm-parse-rec (str)
  "Internal recursive rpm parsing function"
  (when str
    (let ((beg nil)
	  (end nil)
	  (tag nil))
      (if (setq beg (string-match "%{" str)) ;; there is some macro to parse
	  (and (setq end (string-match "}" str))
	       (< beg end)
	       (setq tag (rpm-find-value (substring str (+ beg 2) end)))
	       (rpm-parse-rec (concat (substring str 0 beg) 
				      tag
				      (substring str (1+ end)))))
	;; there is no (more) macro to parse
	(progn str)
	))))



;; ########################## path functions ##############################


;;(setq rpm-dir (expand-file-name rpm-dir))

(setq rpm-dir-build (file-name-as-directory (rpm-eval "%{_builddir}"))) ; .../BUILD
(setq rpm-dir-sources (file-name-as-directory (rpm-eval "%{_sourcedir}"))) ; .../SOURCES
(setq rpm-dir-specs (file-name-as-directory (rpm-eval "%{_specdir}"))) ; .../SPECS
(setq rpm-dir-tmp (file-name-as-directory (rpm-eval "%{tmpdir}")))

;;(make-local-variable 'rpm-dir-buildroot)
;;(setq rpm-dir-buildroot nil)

(defun rpm-dir-buildroot ()
  "Returns parsed Buildroot directory name"
  (rpm-parse 
     ;; either find buildroot
    (replace-in-string 
     (replace-in-string 
      (or (rpm-find-value "Buildroot") (error "Cannot get BuildRoot"))
      "%(id.*)" (getenv "USER")) ; this is hack to speed up parsing if string contains %(id -u -n)
     "%{tmpdir}" (directory-file-name rpm-dir-tmp)) ; this is hack to speed up parsing if string contains %{tmpdir}
    )
  )
    

;;(make-local-variable 'rpm-dir-src)
(defun rpm-dir-src ()
  "Returns parsed name of directory where the build process is run (e.g. .../BUILD/pkg-0.0.2)"
  (rpm-parse
   (or 
    ;; either it is defined by %setup -n <subdirname>
    (save-excursion
       (beginning-of-buffer)
	(if (re-search-forward "^%setup.*[ \t]-n[ \t]" nil t) ; errors off
	    (concat rpm-dir-build 
		    (file-name-as-directory 
		     (rpm-parse
		      (let ((beg (point)))
			(skip-chars-forward "^ \t\n")
			(buffer-substring beg (point))
			))))))
    ;; or is assumed to be %{name}-%{version}
    (file-name-as-directory 
     (concat rpm-dir-build (rpm-find-value "name") "-" (rpm-find-value "version")))
    )))




;; ################## some helpful functions ###################

(defun rpm-find-value (tag)
  "Search specfile for value of given tag, given by preamble field or %define'd."
  "Returns nil if search failed. Give just tag name like 'name' 'version' 'mydef'"
  (if (string= "" tag) 
      nil
    (save-excursion
      ;; e.g find "Name:" "%define macro" "Source1:"
;; TODO searching from beginning (in case version is redefined in package)
      (cond ((re-search-backward (concat "^%define[ \t]+" tag "[ \t]") nil t) ; errors off
	     (forward-char 7)
	     (skip-chars-forward " \t")
	     (forward-char (length tag))
	     (skip-chars-forward " \t")
	     (let ((beg (point)))
	       (skip-chars-forward "^#\t\n") ; stop at tab, nl, or '#' (spaces are sometimes part of macro)
	       (skip-chars-backward " ") ; cut trailing spaces
	       (buffer-substring beg (point))
	       ))
	    ((re-search-backward (concat "^" tag ":[ \t]") nil t)  ;; instead of ((progn (beginning-of-buffer) (re-search-forward (concat "^" tag ":[ \t]") nil t))
	     ;;(forward-char (length tag))
	     (skip-chars-forward "^ \t")
	     (skip-chars-forward " \t")
	     (let ((beg (point)))
	       (skip-chars-forward "^#\t\n")
	       (skip-chars-backward " ") ; cut trailing spaces
	       (buffer-substring beg (point))
	       ))))))


(defun string-before () "String before point"
    (save-excursion
      (let ((here (point)))
      (skip-chars-backward "^ \t\n")
      (setq pattern (buffer-substring (point) here)))))


(defun first-word () "First word in current line as string"
  ;;(interactive "*")
  (save-excursion
    (beginning-of-line)
    (let ((beg (point)))
      (forward-word 1)
      (buffer-substring beg (point))
      )
    )
  )

(defun first-string () "First string in current line"
  ;;(interactive "*")
  (save-excursion
    (beginning-of-line)
    (let ((beg (point)))
      (skip-chars-forward "^ \t\n")
      (buffer-substring beg (point))
      )
    )
  )

(defun first-word-a () "Return first _letters_ of first word in current line as string"
  ;;(interactive "*")
  (save-excursion
    (beginning-of-line)
    (let ((beg (point)))
      (skip-chars-forward "a-zA-Z")
      (buffer-substring beg (point))
      )
    )
  )


(defun current-line () "Return current line before point as string"
  ;;(interactive "*")
  (save-excursion
    (let ((here (point)))
      (beginning-of-line)
      (buffer-substring (point) here)
      )
    )
  )

(defun make-completion-table (strings)
  "Make a completion table from a list of strings."
  (let ((tab nil))
    (while strings
      (setq tab (cons (list (car strings) 0 nil 0) tab))
      (setq strings (cdr strings))
      )
    (progn tab) ;; is this correct??
    )
  )


;(defconst rpm-tags-completion-list
;  '(


;; ##############################################################################

(defconst rpm-macros-completion-list1
  '(
    "%{SOURCE1}"
    "%{SOURCE2}"
    "%{SOURCE3}"
    "%{SOURCE4}"
    "%{SOURCE5}"
    "%{SOURCE6}"
    "%{SOURCE7}"
    "%{SOURCE8}"
    "%{SOURCE9}"
    "%{name}"
    "%{version}"
    "%{release}"
    ;"$RPM_BUILD_ROOT"
    ;"$RPM_OPT_FLAGS"
    )
  "List of some useful macros.")


(defconst rpm-macros-completion-list2
  '(
    "%{_applnkdir}"
    "%{_bindir}"
    "%{_exec_prefix}"
    "%{_datadir}"
    "%{_includedir}"
    "%{_libdir}"
    "%{_libexecdir}"
    "%{_localstatedir}"
    "%{_prefix}"
    "%{_sbindir}"
    "%{_sysconfdir}"
    "%{_sharedstatedir}"
    "%{_mandir}"
    "%{_infodir}"
    "%{__make}"
    )
  "List of macros defining paths. (starting with '%{_')")



;; not used
(defconst rpm-tags-completion-list
  '(
    ;; preamble tags
    "%define "
    "%description"
    "%package "
    ;; prep tags
    "%prep"
    "%setup"
    "%patch"
    ;; build tags
    "%build "
    "%configure "
    "%install "
    ;; scripts tags
    "%post "
    "%postun "
    "%pre "
    "%preun "
    "%clean\nrm -rf $RPM_BUILD_ROOT\n"
    ;; files tags
    "%files\n%defattr(644,root,root,755)\n"
    "%attr "
    "%dir "
    "%doc "
    "%ghost "
    "%config "
    )
  "List which elements are valid tags.")

;; #################################################################

(defconst rpm-preamble-completion-list
  '(
    "Autoreqprov:\t"
    "BuildArch:\t"
    "BuildRequires:\t"
    "BuildRoot:\t%{tmpdir}/%{name}-%{version}-root-%(id -u -n)"
    "Conflicts:\t"
    "Copyright:\t"
    "Distribution:\t"
    "Epoch:\t"
    "ExcludeArch:\t"
    "ExcludeOS:\t"
    "ExclusiveArch:\t"
    "ExclusiveOS:\t"
    "Group:\t\t"
    "Group(pl):\t"
    "Icon:\t\t"
    "License:\t"
    "Name:\t\t"
    "Name(pl):\t"
    "Nopatch:\t"
    "Nosource:\t"
    "Packager:\t"
    "Patch:\t\t"
    "Patch0:\t\t"
    "Patch1:\t\t"
    "Patch2:\t\t"
    "Patch3:\t\t"
    "Patch4:\t\t"
    "Patch5:\t\t"
    "Patch6:\t\t"
    "Patch7:\t\t"
    "Patch8:\t\t"
    "Patch9:\t\t"
    "Prefix:\t\t"
    "Provides:\t"
    "Prereq:\t\t"
    "Obsoletes:\t"
    "Release:\t"
    "Requires:\t"
    "Serial:\t\t"
    "Source:\t\t"
    "Source0:\t"
    "Source1:\t"
    "Source2:\t"
    "Source3:\t"
    "Source4:\t"
    "Source5:\t"
    "Source6:\t"
    "Source7:\t"
    "Source8:\t"
    "Source9:\t"
    "Summary:\t"
    "Summary(pl):\t"
    "URL:\t\t"
    "Vendor:\t\t"
    "Version:\t"
    "%define\t\t"
    "%description" ;; we can start new section too
    )
  "Tags to complete in preamble section.")


(defconst rpm-description-completion-list
  '(
    "%description" ; if we want to add another language description
    "%package"
    "%prep"
    )
  "List which elements are valid tags.")


(defconst rpm-prep-completion-list
  '(
    "%build"
    "%install" ; build section is not mandatory
    "%patch -p1"
    "%patch0 -p1"
    "%patch1 -p1"
    "%patch2 -p1"
    "%patch3 -p1"
    "%patch4 -p1"
    "%patch5 -p1"
    "%patch6 -p1"
    "%patch7 -p1"
    "%patch8 -p1"
    "%patch9 -p1"
    "%setup -q"
    )
  "List which elements are valid tags in %prep section.")


(defconst rpm-build-completion-list
  '(
    "CFLAGS=\"$RPM_OPT_FLAGS\"; export CFLAGS"
    "CXXFLAGS=\"$RPM_OPT_FLAGS\"; export CXXFLAGS"
    "CPPFLAGS=\"$RPM_OPT_FLAGS\"; export CPPFLAGS"
    "FFLAGS=\"$RPM_OPT_FLAGS\"; export FFLAGS"
    "LDFLAGS=\"-s\"; export LDFLAGS"
    "%configure "
    "%install "
    "%post "
    "%postun "
    "%pre "
    "%preun "
    "aclocal "
    "autoconf "
    "autoheader "
    "automake "
    ;;"configure %{_target_patform}"
    "export "
    "gettextize "
    "make "
    )
  "Completion list for %build.")

(defconst rpm-install-completion-list
  '(
    "%clean\nrm -rf $RPM_BUILD_ROOT"
    "%configure "
    "%post "
    "%postun "
    "%pre "
    "%preun "
    "install "
    "gzip -9nf "
    "make install"
    "strip"
    "strip --strip-unneeded $RPM_BUILD_ROOT%{_libdir}/lib*.so.*.*"
    )
  "Completion list for %install.")


(defconst rpm-scripts-completion-list
  '(
    ;;"%clean\nrm -rf $RPM_BUILD_ROOT\n"
    "%files\n%defattr(644,root,root,755)\n"
    "%post "
    "%postun "
    "%pre "
    "%preun "
    "/sbin/ldconfig"
    "[ ! -x /usr/sbin/fix-info-dir ] || /usr/sbin/fix-info-dir -c %{_infodir} >/dev/null 2>&1"
    "%{_sbindir}/fix-sgml-catalog"
    )
  "List which elements are valid tags in scripts.")

(defconst rpm-files-completion-list
  '(
    "%attr(755,root,root) "
    "%config(noreplace) "
    "%config(missingok) "
    "%dir "
    "%doc "
    "%defattr(644,root,root,755) "
    "%files "
    "%ghost "
    ;;"%lang(pl)" ;; to be continued
    "%lang("
    "%verify("
    )
  "List of macros specific to %files section.")





(defconst rpm-groups-completion-list
  '(
    "Applications"
    "Applications/Communications"
    "Applications/Editors"
    "Applications/Editors/Emacs"
    "Applications/Emulators"
    "Applications/Engineering"
    "Applications/Spreadsheets"
    "Applications/Databases"
    "Applications/Graphics"
    "Applications/Networking"
    "Applications/Mail"
    "Applications/Math"
    "Applications/News"
    "Applications/Publishing"
    "Applications/Publishing/TeX"
    "Applications/Publishing/SGML"
    "Applications/Publishing/XML"
    "Applications/Sound"
    "Base"
    "Base/Kernel"
    "Utilities"
    "Utilities/Archiving"
    "Utilities/Console"
    "Utilities/File"
    "Utilities/Printing"
    "Utilities/System"
    "Utilities/Terminal"
    "Utilities/Text"
    "Daemons"
    "Documentation"
    "X11"
    "X11/XFree86"
    "X11/XFree86/Servers"
    "X11/Applications"
    "X11/Applications/Graphics"
    "X11/Applications/Networking"
    "X11/Games"
    "X11/Games/Strategy"
    "X11/Games/Video"
    "X11/Amusements"
    "X11/Utilities"
    "X11/Libraries"
    "X11/Window Managers"
    "Libraries"
    "Networking"
    "Networking/Admin"
    "Networking/Daemons"
    "Networking/News"
    "Networking/Utilities"
    "Development"
    "Development/Debuggers"
    "Development/Libraries"
    "Development/Libraries/Libc"
    "Development/Languages"
    "Development/Languages/Fortran"
    "Development/Languages/Tcl"
    "Development/Building"
    "Development/Version Control"
    "Development/Tools"
    "Shells"
    "Games"
    )
  "List which elements are valid group tags.")



;; mieci ##################################

(defun rpm-filename-completion-table (eltypes)                                                       
  "Make a completion table from a list, ELTYPES, of element types."                                 
  (loop for et in eltypes as name = (sgml-eltype-name et)                                           
        if (boundp et)                                                                              
        collect (cons name name)))





;;; rpm-spec-mode.el ends here
